#include "mocap_inverse_kinematics.h"
#include "SDModel.h"
#include "SDModel_common.h"
#include "sdlib.h"
#include "primus_common.h"
#include <iostream>
#include <vector>

static primus_model_t ik_model;
static SDModel *m = &ik_model;

static bool firsttime = true;

#include <Eigen/Core>
//#include <Eigen/LU>

#include "eiquadprog.hpp"

#define NVARS (N_U+3)	// joint velocities plus translation

// import most common Eigen types  
using namespace Eigen;

class Objective{
public:
	Objective(){
		A.setZero(0,0);
		b.setZero(0);
	}

	Eigen::MatrixXd getQuadraticObj(){
		return A.transpose()*A;
	};

	Eigen::VectorXd getLinearObj(){
		return -b.transpose()*A;
	};

	Eigen::MatrixXd A;
	Eigen::VectorXd b;
	double weight;
	int rows;
	int cols;
};

class multiObjective: public Objective{
public:
	multiObjective(int _NVARS){
		cols = _NVARS;
	};

	void addObjective(Eigen::MatrixXd &_A, Eigen::VectorXd &_b, double _weight){
		A.conservativeResize(A.rows()+_A.rows(),cols);
		A.bottomRows(_A.rows()) = _weight*_A;
		b.conservativeResize(b.rows()+_b.rows());
		b.bottomRows(_b.rows()) = _weight*_b;
		rows+=_A.rows();
	}

private:

};

void mocap_markers_kinematics(BEN_COMMON_t *p){
	double temp[6];
	//double Jtemp[6*N_U];
	int i,j;
	set_state(m,p->ik_cur.rob.state);
	for(i=0;i<N_MARKERS;i++){
		double btj[3];

		sdvset(0,0,0,p->mocap_cur.pos[i]);
		

		for(j=0;j<N_J;j++){
			int info[50];
			int slider[6];
			m->sdjnt(m->joint(j),info,slider);
			if(info[3]==MARKER_BODIES[i]) break;
		}

		if(j<N_J){
			// get position of joint
			int joint = m->joint(j);
			m->sdgetbtj(joint,btj);

		}else{
			// get position of body
			sdvset(0,0,0,btj);
		}

		body_position(m,NULL,MARKER_BODIES[i],btj,temp,NULL);

		sdvcopy(temp,p->mocap_cur.pos[i]);
	}	
}

void mocap_markers_kinematics_with_Jac(BEN_COMMON_t *p, MatrixXd &ARef, VectorXd &bRef){
	double temp[6];
	double Jtemp[6*N_U];
	int i,j;
	set_state(m,p->ik_cur.rob.state);
	for(i=0;i<N_MARKERS;i++){
		double btj[3];

		sdvset(0,0,0,p->mocap_cur.pos[i]);
		

		for(j=0;j<N_J;j++){
			int info[50];
			int slider[6];
			m->sdjnt(m->joint(j),info,slider);
			if(info[3]==MARKER_BODIES[i]) break;
		}

		if(j<N_J){
			// get position of joint
			int joint = m->joint(j);
			m->sdgetbtj(joint,btj);

		}else{
			// get position of body
			sdvset(0,0,0,btj);
		}

		body_position(m,NULL,MARKER_BODIES[i],btj,temp,Jtemp);
	
		sdvcopy(temp,p->mocap_cur.pos[i]);

		ARef.block(3*i,0,3,N_U) = MatrixXd::Map(Jtemp,6,N_U).block(0,0,3,N_U);
		bRef.segment(3*i,3) = VectorXd::Map(p->mocap_des.vel[i],3) + p->IK_POS_RATE*(VectorXd::Map(p->mocap_des.pos[i],3) + VectorXd::Map(p->mocap_offset,3) - VectorXd::Map(p->mocap_cur.pos[i],3));
	}	
}

void mocap_ik_init(void){
	m->sdinit();
}


void mocap_inverse_kinematics(BEN_COMMON_t *p){
/* PRIMUS_t is just a big structure of common variables */

	int i,j;
	multiObjective obj = multiObjective(NVARS);

	double PL[6],PR[6],JL[N_U*6]={0.0},JR[N_U*6]={0.0};	
	double PT[6],JT[N_U*6]={0.0};
	double zero[3]={0.0};
	double com[3],JC[3*N_U]={0.0};
	double temp[6];
	
	double DT = 0.0025;	// timestep (or gradient descent step size)

	p->IK_POS_RATE = 0.1;

	if (p->count==0){
		m->sdinit();
		for(i=0;i<N_Q;i++)	p->ik_cur.rob.state[i] = p->ref.rob.state[i];	
		primus_state_init(p->ik_cur.rob.state);
		primus_forward_kinematics(m,p->ik_cur.rob.state,p->ik_des.sim.com,p->ik_des.sim.foot[0],p->ik_des.sim.foot[1],p->ik_des.sim.torso);
		p->ik_des.sim.foot[0][2] = 0.0;
		p->ik_des.sim.foot[1][2] = 0.0;
		sdvset(0,0,0,&p->ik_des.sim.foot[0][3]);
		sdvset(0,0,0,&p->ik_des.sim.foot[1][3]);

		for(i=0;i<3;i++)	p->mocap_offset[i] = 0.0;
	}

	set_state(m,p->ik_cur.rob.state);
	
	com_state(m,p->ik_cur.sim.com,p->ik_cur.sim.comd);
	//sdvsub(p->ik_cur.sim.com,p->com_offset,p->ik_cur.sim.com);

	primus_feet(m,p->ik_cur.sim.foot[0],JL,p->ik_cur.sim.foot[1],JR);

	m->sdgetbtj(PRIMUS_RIGHT_FOOT,temp);
	m->sdvel(PRIMUS_RIGHT_FOOT,temp,p->ik_cur.sim.footd[0]);
	m->sdgetbtj(PRIMUS_LEFT_FOOT,temp);
	m->sdvel(PRIMUS_LEFT_FOOT,temp,p->ik_cur.sim.footd[1]);

	primus_hip(m,temp,JT);
	sdvcopy(&temp[3],p->ik_cur.sim.torso);

	com_jacobian(m,JC);

	MatrixXd JacC(MatrixXd::Map(JC,3,N_U));
	MatrixXd JacT(MatrixXd::Map(JT,6,N_U));

	// com objective
	/***************************
	 * ACOM = [JC];
	 * bCOM = comd_d
	 ****************************/
	MatrixXd ACOM(MatrixXd::Zero(3,NVARS));	
	VectorXd bCOM(3);
	
	ACOM.block(0,0,3,N_U) = JacC;
	bCOM.segment(0,3) = VectorXd::Map(p->ik_des.sim.comd,3) + p->IK_POS_RATE*(VectorXd::Map(p->ik_des.sim.com,3) - VectorXd::Map(p->ik_cur.sim.com,3));

	//=== feet constraints / accelerations
	/***************************
	 * AFeet = [JL;
	 *          JR]
	 * bFeet = [footd_d[LEFT];
	 *          footd_d[RIGHT]];
	 ****************************/

	MatrixXd AFeet(MatrixXd::Zero(12,NVARS));
	VectorXd bFeet(12);
	
	AFeet.block(0,0,6,N_U) = (MatrixXd::Map(JL,6,N_U));
	AFeet.block(6,0,6,N_U) = (MatrixXd::Map(JR,6,N_U));

	bFeet.segment(0,3) = VectorXd::Map(p->ik_des.sim.footd[0],3) + p->IK_POS_RATE*(VectorXd::Map(p->ik_des.sim.foot[0],3) - VectorXd::Map(p->ik_cur.sim.foot[0],3));
	bFeet.segment(3,3) = p->IK_POS_RATE*( - VectorXd::Map(&p->ik_cur.sim.foot[0][3],3));

	bFeet.segment(6,3) = VectorXd::Map(p->ik_des.sim.footd[1],3) + p->IK_POS_RATE*(VectorXd::Map(p->ik_des.sim.foot[1],3) - VectorXd::Map(p->ik_cur.sim.foot[1],3));
	bFeet.segment(9,3) = p->IK_POS_RATE*( - VectorXd::Map(&p->ik_cur.sim.foot[1][3],3));


	//======= desired torso angle
	 /***************************
	 * ATorso = [JT]
	 * bTorso = [torsod_d];
	 ****************************/
	MatrixXd ATorso(MatrixXd::Zero(3,NVARS));
	VectorXd bTorso(VectorXd::Zero(3));

	ATorso.block(0,0,3,N_U) = JacT.block(3,0,3,N_U);
	bTorso = VectorXd::Map(p->ik_des.sim.torsod,3) + p->IK_POS_RATE*(VectorXd::Map(p->ik_des.sim.torso,3) - VectorXd::Map(p->ik_cur.sim.torso,3));

	//======= minimize joint velocities
	 /***************************
	 * AJoint = [I]
	 * bJoint = [0];
	 ****************************/
	MatrixXd AJoint(MatrixXd::Zero(N_U,NVARS));
	VectorXd bJoint(VectorXd::Zero(N_U));
	
	AJoint.block(0,0,N_U,N_U) = MatrixXd::Identity(N_U,N_U);
	
	//======== motion capture reference
	/***************************
	 * ARef = [J_i]
	 * bRef = [Pid_ref + kp(Pi_ref - Pi)];
	 ****************************/
	double Jtemp[6*N_U];
	MatrixXd ARef(MatrixXd::Zero(3*N_MARKERS,NVARS));
	VectorXd bRef(VectorXd::Zero(3*N_MARKERS));

	mocap_markers_kinematics_with_Jac(p,ARef,bRef);

	for(i=0;i<N_MARKERS;i++){
		ARef.block(3*i,N_U,3,3) = MatrixXd::Identity(3,3);	// adds a translation
	}

	//=======
	
	MatrixXd CE(NVARS,0);
	VectorXd ce0(1); // can't set vectors to zero length

	//=======
	
	MatrixXd CI(MatrixXd::Zero(NVARS,2));
	VectorXd ci0(VectorXd::Zero(2));

	// Example KFE joint limit constraint
	// KFE + T*KFEd >= 0.5	==>  T*KFEd + KFE - 0.5 >= 0

	CI(18,0)= DT;	ci0(0)	= p->ik_cur.rob.state[18] - 0.5;	// robot straight knee
	CI(25,1)= DT;	ci0(1)	= p->ik_cur.rob.state[25] - 0.5;	// robot straight knee

//	//========
//	
	obj.addObjective(AFeet,bFeet,1.0);
	obj.addObjective((MatrixXd)ACOM.topRows(2),(VectorXd)bCOM.topRows(2),1.0);
	//obj.addObjective((MatrixXd)ACOM.bottomRows(1),(VectorXd)bCOM.bottomRows(1),0.001);
	obj.addObjective((MatrixXd)ATorso.topRows(2),(VectorXd)bTorso.topRows(2),0.1);
	obj.addObjective(ARef,bRef,0.001);
	obj.addObjective(AJoint,bJoint,0.0001);

	// costs are weighted sum of torque, constaints and com objectives
	MatrixXd G(NVARS,NVARS);
	VectorXd g0(NVARS);
	
	G = obj.getQuadraticObj();
	g0 = obj.getLinearObj();

	//====

	VectorXd x(NVARS);

	solve_quadprog(G, g0, CE, ce0, CI, ci0, x);

	VectorXd::Map(p->ik_cur.rob.stated,N_U) = x.segment(0,N_U);

	for(i=0;i<m->NU();i++)	p->ik_cur.rob.state[N_Q+i] = p->ik_cur.rob.stated[i];

	double delta = DT;
	double base_q[4];
	double base_qd[4];

	Vector3d::Map(p->mocap_offset) += delta*x.tail(3);

	base_q[0] = p->ik_cur.rob.state[3];
	base_q[1] = p->ik_cur.rob.state[4];
	base_q[2] = p->ik_cur.rob.state[5];
	base_q[3] = p->ik_cur.rob.state[m->NU()];

	p->ik_cur.rob.state[0] += delta*p->ik_cur.rob.stated[0];
	p->ik_cur.rob.state[1] += delta*p->ik_cur.rob.stated[1];
	p->ik_cur.rob.state[2] += delta*p->ik_cur.rob.stated[2];

	for(i=0;i<m->NJNT();i++)	p->ik_cur.rob.state[6+i] += delta*p->ik_cur.rob.stated[6+i];

	sdvmul(base_q[3],&p->ik_cur.rob.stated[3],base_qd);
	sdvcross(base_q,&p->ik_cur.rob.stated[3],temp);
	sdvsub(base_qd,temp,base_qd);
	sdvmul(0.5,base_qd,base_qd);
	base_qd[3]= -0.5*sdvdot(base_q,&p->ik_cur.rob.stated[3]);

	p->ik_cur.rob.state[3] += delta*base_qd[0];
	p->ik_cur.rob.state[4] += delta*base_qd[1];
	p->ik_cur.rob.state[5] += delta*base_qd[2];
	p->ik_cur.rob.state[m->NU()] += delta*base_qd[3];

	// compute the resulting IK velocities
	Vector3d::Map(p->ik_cur.sim.comd) = JacC*x.segment(0,N_U);
	VectorXd::Map(p->ik_cur.sim.footd[0],6) = AFeet.block(0,0,6,N_U)*x.segment(0,N_U);
	VectorXd::Map(p->ik_cur.sim.footd[1],6) = AFeet.block(6,0,6,N_U)*x.segment(0,N_U);
		
	if(p->count%100==0){
		std::cout << "offset: " << Vector3d::Map(p->mocap_offset).transpose() << "\n" << std::endl;

	}
	//std::cout << "T: " << torque.transpose() << "\n" << std::endl;
	//std::cout << "F: " << force.transpose() << "\n\n" << std::endl;
}