#ifndef MOCAP_INVERSE_KINEMATICS2_H
#define MOCAP_INVERSE_KINEMATICS2_H


#include "primus_common.h"



enum MARKER_IDX{
	RANK = 0,
	LANK,
	RKNE,
	LKNE,
	RBWT,
	LBWT,
	STRN,
	CLAV,
	RSHO,
	LSHO,
	RELB,
	LELB,
	RWRA,
	LWRA,
	N_MARKERS
};


static int MARKER_BODIES[N_MARKERS] = {
	33,
	26,
	31,
	24,
	28,
	21,
	3,
	4,
	7,
	14,
	9,
	16,
	11,
	18
};

typedef struct MOCAP_t{
	double pos[N_MARKERS][3];
	double vel[N_MARKERS][3];
} MOCAP_t;


typedef struct SIMPLE_STATE_t{
	double com[3];
	double comd[3];
	double comdd[3];

	double foot[2][6];
	double footd[2][6];
	double footdd[2][6];

	double torso[3];
	double torsod[3];
	double torsodd[3];
} SIMPLE_STATE_t;

typedef struct ROBOT_STATE_t{
	double state[N_U+N_Q];
	double stated[N_U+N_U];
	double torque[N_J];
	double foot_force[2][6];
} ROBOT_STATE_t;

typedef struct STATE_t{
	SIMPLE_STATE_t sim;
	ROBOT_STATE_t rob;
} STATE_t;

typedef struct BEN_COMMON_t{
	STATE_t ik_cur;
	STATE_t ik_des;
	STATE_t ref;
	STATE_t home;

	MOCAP_t mocap_cur;
	MOCAP_t mocap_des;

	double mocap_offset[3];

	int count;

	double IK_POS_RATE;

} BEN_COMMON_t;

void mocap_ik_init(void);
void mocap_inverse_kinematics(BEN_COMMON_t *p);
void mocap_markers_kinematics(BEN_COMMON_t *p);

#endif