#include "mocap_inverse_kinematics.h"
#include "SDModel.h"
#include "SDModel_common.h"
#include "sdlib.h"
#include "primus_common.h"
#include "mrdplot.h"
#include "trajectory.h"
#include <iostream>
#include <vector>

//static primus_model_t ik_model;
//static SDModel *m = &ik_model;

static bool firsttime = true;

#include <Eigen/Core>

static BEN_COMMON_t common;

//static void primus_markers(BEN_COMMON_t *p){
//	int i,j;
//	double Jtemp[6*N_U];
//	double temp[6];
//	double zero[3]={0};
//
//	for(i=0;i<N_MARKERS;i++){
//		double btj[3];
//
//		sdvset(0,0,0,p->mocap_cur.pos[i]);
//		
//		printf("%i %i - ",i,MARKER_BODIES[i]);
//
//		for(j=0;j<N_J;j++){
//			int info[50];
//			int slider[6];
//			m->sdjnt(m->joint(j),info,slider);
//			if(info[3]==MARKER_BODIES[i]) break;
//		}
//
//		if(j<N_J){
//			// get position of joint
//			printf("%i -",m->joint(j));
//			int joint = m->joint(j);
//			m->sdgetbtj(joint,btj);
//		}else{
//			// get position of body
//			sdvset(0,0,0,btj);
//		}
//		body_position(m,NULL,MARKER_BODIES[i],btj,temp,Jtemp);
//
//		sdvcopy(temp,p->mocap_cur.pos[i]);
//		printf(" % f % f % f \n",p->mocap_cur.pos[i][0],p->mocap_cur.pos[i][1],p->mocap_cur.pos[i][2]);
//
//		printf("\n");
//	}
//}

//#define MAX_FRAMES 5000
//static double mocapdata[MAX_FRAMES][2*3*N_MARKERS];
//static int mocapframes;
static double MOCAP_SPEED = 1.0;
static double MOCAP_TRAJ_LENGTH;
static bool mocap_allocated = false;
static DOFS mocapTraj[3*N_MARKERS];
MRDPLOT_DATA *mrdMocap;

static int load_mocap(){

	//========== initialization
	if(!mocap_allocated){

		mrdMocap = read_mrdplot("20_01.mrd");
	
		//==========	

		printf("mrdMocap allocated\n");
	
		mocap_allocated = true;		
	}

	return 1;
}

static int load_mocap_trajectory(BEN_COMMON_t *p){
		int i,j,k;
		int len = mrdMocap->n_points;
		int nchan = mrdMocap->n_channels;
		double freq = mrdMocap->frequency;
		MOCAP_TRAJ_LENGTH = 6.0+(len/freq)/MOCAP_SPEED;
		float data;

		clear_all_knots(mocapTraj,3*N_MARKERS);
		init_trajectory(mocapTraj,3*N_MARKERS);
		
		if(!mocap_allocated) return 0;

		for(i=0;i<N_MARKERS;i++){
			for(k=0;k<3;k++){
			
				// load in the default pose into the trajectory
				add_knot_point( mocapTraj, 3*i+k, QUINTIC_SPLINE, 0, p->mocap_cur.pos[i][k], 0.0, 0.0 );


				int idx = 3*i+k;
		
				for(j=0;j<len;j++){		
					data = mrdMocap->data[j*nchan+idx];
					add_knot_point( mocapTraj, 3*i+k, LINEAR_SPLINE, 3.0+j/freq/MOCAP_SPEED, (double)data, 0.0, 0.0 );
				}
			
				// load in the default pose into the trajectory
				add_knot_point( mocapTraj, 3*i+k, QUINTIC_SPLINE, MOCAP_TRAJ_LENGTH, p->mocap_cur.pos[i][k], 0.0, 0.0 );
			}
		}

		printf("Mocap Trajectory loaded speed = %f length = %f\n",MOCAP_SPEED,MOCAP_TRAJ_LENGTH);

		return mrdMocap->n_points;
}


void main(){
	int i,k,n;
	double temp;

	BEN_COMMON_t *p = &common;
	
	mocap_ik_init();

	for(i=0;i<N_Q;i++)	p->ik_des.rob.state[i] = p->ik_cur.rob.state[i] = 0.0;

	primus_state_init(p->ik_cur.rob.state);
	primus_state_init(p->ik_des.rob.state);


	//=================

	load_mocap();

	//=================
	// load current locations
	mocap_markers_kinematics(p);
	
	//=================
	// load trajectory
	
	int nframes = load_mocap_trajectory(p);

	//=================

	double traj_time = 0.0;

	FILE *fdes = fopen("des","w");
	FILE *fcur = fopen("cur","w");
	FILE *fjnt = fopen("jnt","w");

	for(n=0;n*0.0025<MOCAP_TRAJ_LENGTH;n++){

		
		traj_time = n*0.0025;

		fprintf(fdes,"% f ",traj_time);
		fprintf(fcur,"% f ",traj_time);
		fprintf(fjnt,"% f ",traj_time);

		// look up desired locations
		for(i=0;i<N_MARKERS;i++){
			for(k=0;k<3;k++){
				if(!lookup_spline3_alt( mocapTraj, 3*i+k, traj_time, NULL, &p->mocap_des.pos[i][k], &p->mocap_des.vel[i][k], &temp)){
					p->mocap_des.pos[i][k] = p->mocap_cur.pos[i][k];
					p->mocap_des.vel[i][k] = 0;
				}

				p->mocap_des.pos[i][k] -= p->mocap_offset[k];

				fprintf(fdes,"% f ",p->mocap_des.pos[i][k]);
				fprintf(fcur,"% f ",p->mocap_cur.pos[i][k]);
			}
		}

		//==================
		mocap_inverse_kinematics(p);

		

		for(i=0;i<N_Q;i++)	fprintf(fjnt,"% f ",p->ik_cur.rob.state[i]);

		fprintf(fdes,"\n");
		fprintf(fcur,"\n");
		fprintf(fjnt,"\n");

		p->count++;
	}

	fclose(fdes);
	fclose(fcur);
	fclose(fjnt);


	printf("\n\nFinished - Press Enter to end program\n");
	std::cin.get();
}