#include "SDModel.h"
#include "SDModel_common.h"
#include "sdlib.h"
#include "primus_common.h"
//#include "ben_common.h"
//#include "gsl_util.h"
//#include <gsl/gsl_linalg.h>
//#include <gsl/gsl_blas.h>



static double right_offset[3] = {0.0525, -0.011, -0.05};      
static double left_offset[3] = {0.0525, 0.011, -0.05};
static double hip_offset[3] = {0.0527,-0.011,0.0058};
//static double hip_offset[3] = {-0.043829,-0.008545,0.116};

void primus_feet(SDModel *m, double *fL, double *JfL, double *fR, double *JfR){
    double temp[3];
    // calculate cof of left foot
    m->sdgetbtj(PRIMUS_LEFT_FOOT,temp);
    sdvadd(temp,left_offset,temp);
    body_position(m, NULL, PRIMUS_LEFT_FOOT, temp, fL, JfL);

    // calculate cof of right foot
    m->sdgetbtj(PRIMUS_RIGHT_FOOT,temp);
    sdvadd(temp,right_offset,temp);
    body_position(m, NULL, PRIMUS_RIGHT_FOOT, temp, fR, JfR);     
}

void primus_ankles(SDModel *m, double *fL, double *JfL, double *fR, double *JfR){
    double temp[3];
	double zero[3]={0.0};
    // calculate center of left ankle
    m->sdgetbtj(PRIMUS_LEFT_FOOT,temp);
    sdvadd(temp,zero,temp);
    body_position(m, NULL, PRIMUS_LEFT_FOOT, temp, fL, JfL);

    // calculate center of right ankle
    m->sdgetbtj(PRIMUS_RIGHT_FOOT,temp);
    sdvadd(temp,zero,temp);
    body_position(m, NULL, PRIMUS_RIGHT_FOOT, temp, fR, JfR);   

	// add offsets as if the feet are flat
	sdvadd(fL,left_offset,fL);
	sdvadd(fR,right_offset,fR);
}

void primus_hip(SDModel *m, double *ph, double *Jh){
	body_position(m, NULL, PRIMUS_TORSO, hip_offset, ph, Jh);     
}

void primus_feetcorners(SDModel *m, double *fl, double *Jl, double *fr, double *Jr){
    double temp[3],temp2[3],temp3[3];
    double length = 0.25;
    double width = 0.12;
	double *lin;
    int i,j;    
    double offset[4][3] = {{length/2.0,width/2.0,0.0},
                           {length/2.0,-width/2.0,0.0},
                           {-length/2.0,-width/2.0,0.0},
                           {-length/2.0,width/2.0,0.0}};
    // calculate cof of left foot
    m->sdgetbtj(PRIMUS_LEFT_FOOT,temp);
    sdvadd(temp,left_offset,temp);
    for(i=0;i<4;i++){
        sdvadd(temp,offset[i],temp2);
		m->sdpos(PRIMUS_LEFT_FOOT,temp2,&fl[3*i]);
        // if (Jl!=NULL){
			// for(j=0;j<m->NU();j++){
				// lin = &Jl[3*4*i + 3*j]
				// m->sdrel2cart(i,PRIMUS_LEFT_FOOT,temp2,&Jl[6*i+0],temp3);
				// m->sdtrans(PRIMUS_LEFT_FOOT,lin,-1,lin);
			// }
		// }
    }

    // calculate cof of right foot
    m->sdgetbtj(PRIMUS_RIGHT_FOOT,temp);
    sdvadd(temp,right_offset,temp);
    for(i=0;i<4;i++){
        sdvadd(temp,offset[i],temp2);
		m->sdpos(PRIMUS_RIGHT_FOOT,temp2,&fr[3*i]);
        // if (Jl!=NULL){
			// for(j=0;j<m->NU();j++){
				// m->sdrel2cart(i,PRIMUS_RIGHT_FOOT,temp2,&Jr[6*i+0],temp3);
				// m->sdtrans(PRIMUS_RIGHT_FOOT,lin,-1,lin);
			// }
		// }
    }
}

void primus_com_wrt_feet(SDModel *m, double *cl, double *Jcl, double *cr, double *Jcr){
    double pl[6],pr[6];
    double temp[6];
    primus_feet(m, pl, NULL, pr, NULL);
    //com_state(m,cl,NULL);
	primus_hip(m,temp,NULL);
	sdvcopy(temp,cl);
	sdvcopy(cl,cr);
	sdvsub(cl,pl,cl);
	sdvsub(cr,pr,cr);
	m->sdtrans(-1,cl,PRIMUS_LEFT_FOOT,cl);
	m->sdtrans(-1,cr,PRIMUS_RIGHT_FOOT,cr);
}

void primus_com_wrt_ankles(SDModel *m, double *cl, double *cr){
    double pl[6],pr[6];
    primus_ankles(m, pl, NULL, pr, NULL);
    com_state(m,cl,NULL);
	sdvcopy(cl,cr);
	sdvsub(cl,pl,cl);
	sdvsub(cr,pr,cr);

}

void primus_hose_contact(SDModel *m, double *p, double *J){
	double temp[3];
	double offset[3] = {-0.12,0.0,0.0};
	double itj1[3],itj2[3];	
	m->sdgetitj(13,itj1);
	m->sdgetitj(20,itj2);
	//m->sdgetitj(1,itj1);
	//m->sdgetitj(8,itj2);
	sdvadd(itj1,itj2,temp);
	sdvaxpy(-0.5,temp,offset,temp);
	body_position(m,NULL,0,temp,p,J);
}



Primus::Primus(SDModel *model){
	int i;
	m = model;
	JL = new double[m->NU()*6];
	JR = new double[m->NU()*6];
	state = new double [m->NQ()+m->NU()];
	des_state = new double [m->NQ()+m->NU()];
	for(i=0;i<m->NQ()+m->NU();i++){
		state[i] = 0.0;
		des_state[i] = 0.0;
	}
	//gsl_JL = gsl_matrix_view_array(JL,m->NU(),6);
	//gsl_JR = gsl_matrix_view_array(JR,m->NU(),6);
	//gsl_J = gsl_matrix_alloc(m->NU(),12);
}

Primus::~Primus(){
	delete JL;
	delete JR;
	delete state;
	delete des_state;
	//gsl_matrix_free(gsl_J);
}


void primus_forward_kinematics(SDModel *m,
							   double *state,
							   double *com,
							   double *pL,
							   double *pR,
							   double *T){
    double temp[6],Jtemp[6*N_U];
	set_state(m,state);
	primus_feet(m, pL, NULL, pR, NULL);	
	body_position(m, NULL, PRIMUS_TORSO, hip_offset, temp, Jtemp);
	//primus_hip(m,temp,NULL);
#ifdef USE_HIP_AS_COM
	sdvcopy(&temp[0],com);
#else
	com_state(m,com,NULL);
#endif
	sdvcopy(&temp[3],&T[0]);	// torso rotation

	//sdvsub(com,primusVar.com_offset,com);
}

void primus_state_init(double *state){
	static primus_model_t model;
	static int firsttime = 1;
	if(firsttime){
		model.sdinit();
		firsttime=0;
	}
	double temp[6];
	double foot[2][6];
	int i;
	for(i=0;i<6;i++)	state[i] = 0.0;
	state[N_U] = 1.0;
	primus_forward_kinematics(&model,state,temp,foot[0],foot[1],temp);
	state[0] -= 0.5*(foot[0][0]+foot[1][0]);
	state[1] -= 0.5*(foot[0][1]+foot[1][1]);
	state[2] -= 0.5*(foot[0][2]+foot[1][2]);
	for(i=0;i<N_U;i++)	state[N_Q+i] = 0.0;
}