#ifndef PRIMUS_COMMON
#define PRIMUS_COMMON

#include "SDModel.h"
#include "primus_fast4.h"
//#include <gsl/gsl_linalg.h>

#define N_U	29
#define N_Q 30
#define N_J	23

#define PRIMUS_LEFT_FOOT 26
#define PRIMUS_RIGHT_FOOT 33
#define PRIMUS_TORSO 3

#define primus_model_t primus_fast4

class Primus{
public:
	Primus(SDModel*model);
	~Primus();
	SDModel *m;
	double *JL,*JR;
	double PL[6],PR[6];
	double *state,*des_state;
	//gsl_matrix_view gsl_JL,gsl_JR;
	//gsl_matrix *gsl_J;
};



//-------- requires call to set_state prior to exec --------//

// returns centers of feet and orientation (and jacobian)
void primus_feet(SDModel *m, double *fl, double *Jl, double *fr, double *Jr);

// returns ankle joint positions and feet orientation (and jacobian)
void primus_ankles(SDModel *m, double *fl, double *Jl, double *fr, double *Jr);

// returns hip position and orientation (and jacobian)
void primus_hip(SDModel *m, double *ph, double *Jh);

// returns vector of 3D poses of feet corners (and jacobian)
void primus_feetcorners(SDModel *m, double *fl, double *Jl, double *fr, double *Jr);

// COM and velocity w.r.t. each foot
void primus_com_wrt_feet(SDModel *m, double *cl, double *Jcl, double *cr, double *Jcr);

// COM and velocity w.r.t. each ankle
void primus_com_wrt_ankles(SDModel *m, double *cl, double *cr);

// position of point where hoses contact the robot
void primus_hose_contact(SDModel *m, double *p, double *J);

void primus_forward_kinematics(SDModel *m,
							   double *state,
							   double *com,
							   double *pL,
							   double *pR,
							   double *T);

// determines base state so that feet are on the ground
void primus_state_init(double *state);

//// gravity compensation inverse dynamics
//void primus_gravcomp(SDModel *m, double *X, double *U);

void primus_com_ik(SDModel *m, double *state, double *pLd_des, double *pRd_des, double *dpdes, double *dTdes, double *dq);

void run_inverse_kinematics(SDModel *m, 
							double *virt_state, 
							double *virt_com_d,
							double *virt_comd_d, 
							double *virt_pL_d,
							double *virt_pLd_d, 
							double *virt_pR_d,
							double *virt_pRd_d,
							double *virt_torso_d,
							double *virt_torsod_d,
							double delta,
							double tol);

//-------- performs multiple calls to set_state ------------//



//-------- requires call to sdderiv prior to exec ----------//


#endif